/*
	Copyright (C) 2006-2008 Christopher E. Miller
	
	This software is provided 'as-is', without any express or implied
	warranty.  In no event will the authors be held liable for any damages
	arising from the use of this software.
	
	Permission is granted to anyone to use this software for any purpose,
	including commercial applications, and to alter it and redistribute it
	freely, subject to the following restrictions:
	
	1. The origin of this software must not be misrepresented; you must not
	   claim that you wrote the original software. If you use this software
	   in a product, an acknowledgment in the product documentation would be
	   appreciated but is not required.
	2. Altered source versions must be plainly marked as such, and must not be
	   misrepresented as being the original software.
	3. This notice may not be removed or altered from any source distribution.
*/


/**
	Text and strings in D; $(LINK2 http://www.dprogramming.com/mtext.php, _mtext 2.0)
	$(BR) For both Phobos and Tango.
	Trademarks: Unicode(tm) is a trademark of Unicode, Inc.
**/
module mtext;


version(Tango)
{
	import tango.text.convert.Utf;
}
else
{
	import std.utf;
}


version(MSTRING_DISALLOW_FOREACH_MODIFY)
{
}
else
{
	version = MSTRING_ALLOW_FOREACH_MODIFY;
}

// version = MSTRING_CHAR_ONLY;


/**
	Unicode _string structure that stores strings in a format that is easiest to index, slice, and access UTF-32 (dchar) code points,
	as well as attempts to be space efficient.
	The size of this structure is guaranteed to be the size of char[] (_mstring.sizeof == (char[]).sizeof).
	This template allows an allocator.
**/
/// $(B Use the "mstring" type for the default string type.)
struct MString(StringAllocator)
{
	/// This string's allocator.
	alias StringAllocator Allocator;
	
	alias MString mstring;
	
	
	/// Property: get the _length of the string.
	size_t length() // getter
	{
		return _len & LENGTH_MASK;
	}
	
	
	/// Property: get the _type of the underlying array of characters. May be void if not initialized.
	TypeInfo type() // getter
	{
		switch(_len & TYPE_MASK)
		{
			case TYPE_UTF8:
				return typeid(char[]);
			case TYPE_UTF16:
				return typeid(wchar[]);
			case TYPE_UTF32:
				return typeid(dchar[]);
			default: // No type..
				return typeid(void); // ?
		}
	}
	
	
	// set forceRefer to true to force reference
	// returns true if this has its own memory that needs freeing.
	private bool _setr(char[] s, bool forceRefer)
	{
		if(s.length > MAX_LENGTH)
			errortoolong();
		
		size_t sct = TYPE_UTF8;
		
		foreach(dchar ch; s)
		{
			size_t x;
			x = smallestCharType(ch);
			if(x > sct)
			{
				sct = x;
				if(TYPE_UTF32 == x)
					break;
			}
		}
		
		switch(sct)
		{
			case TYPE_UTF32:
				{
					dchar[] ds;
					//ds = std.utf.toUTF32(s);
					ds = _mtoUTFN!(char, dchar)(s);
					_dptr = ds.ptr;
					_len = ds.length;
					_len |= TYPE_UTF32;
				}
				break;
			
			case TYPE_UTF16:
				{
					wchar[] ws;
					//ws = std.utf.toUTF16(s);
					ws = _mtoUTFN!(char, wchar)(s);
					_wptr = ws.ptr;
					_len = ws.length;
					_len |= TYPE_UTF16;
				}
				break;
			
			default:
				if(forceRefer)
					s = Allocator.reference(s);
				_cptr = s.ptr;
				_len = s.length;
				_len |= TYPE_UTF8;
				if(!forceRefer)
					return false;
		}
		return true;
	}
	
	private void _set(char[] s)
	{
		auto r = _setr(s, true);
		assert(r);
	}
	
	// set forceRefer to true to force reference
	// returns true if this has its own memory that needs freeing.
	private bool _setr(wchar[] s, bool forceRefer)
	{
		if(s.length > MAX_LENGTH)
			errortoolong();
		
		size_t sct = TYPE_UTF16;
		
		foreach(dchar ch; s)
		{
			size_t x;
			x = smallestCharType(ch);
			if(x > sct)
			{
				sct = x;
				//if(TYPE_UTF32 == x)
				assert(TYPE_UTF32 == x);
					break;
			}
		}
		
		switch(sct)
		{
			case TYPE_UTF32:
				{
					dchar[] ds;
					//ds = std.utf.toUTF32(s);
					ds = _mtoUTFN!(wchar, dchar)(s);
					_dptr = ds.ptr;
					_len = ds.length;
					_len |= TYPE_UTF32;
				}
				break;
			
			default:
				if(forceRefer)
					s = Allocator.reference(s);
				_wptr = s.ptr;
				_len = s.length;
				_len |= TYPE_UTF16;
				if(!forceRefer)
					return false;
		}
		return true;
	}
	
	private void _set(wchar[] s)
	{
		auto r = _setr(s, true);
		assert(r);
	}
	
	// set forceRefer to true to force reference
	// returns true if this has its own memory that needs freeing.
	private bool _setr(dchar[] s, bool forceRefer)
	{
		if(s.length > MAX_LENGTH)
			errortoolong();
		if(forceRefer)
			s = Allocator.reference(s);
		_dptr = s.ptr;
		_len = s.length;
		_len |= TYPE_UTF32;
		if(!forceRefer)
			return false;
		return true;
	}
	
	private void _set(dchar[] s)
	{
		auto r = _setr(s, true);
		assert(r);
	}
	
	
	// Only releases based on _setr return value.
	private void _unsetr(bool setrResult)
	{
		if(setrResult)
		{
			release();
			return;
		}
		_len = 0;
		_ptr = null;
	}
	
	
	/**
		Create a new instance of mstring.
		The _string passed in may be used as-is, and any string modifying functions may modify the original _string memory.
	**/
	static mstring opCall(char[] s)
	{
		mstring result;
		result._set(s);
		return result;
	}
	
	version(MSTRING_CHAR_ONLY)
	{
	}
	else
	{
		/// ditto
		static mstring opCall(wchar[] s)
		{
			mstring result;
			result._set(s);
			return result;
		}
		
		/// ditto
		static mstring opCall(dchar[] s)
		{
			mstring result;
			result._set(s);
			return result;
		}
	}
	
	
	/// Shortcut for a null, and thus empty, _string: mstring()
	static mstring opCall()
	{
		mstring result; // null AND empty!
		return result;
	}
	
	
	/// Assignment operator.
	void opAssign(char[] s)
	{
		release();
		_set(s);
	}
	
	version(MSTRING_CHAR_ONLY)
	{
	}
	else
	{
		/// ditto
		void opAssign(wchar[] s)
		{
			release();
			_set(s);
		}
		
		/// ditto
		void opAssign(dchar[] s)
		{
			release();
			_set(s);
		}
	}
	
	
	/// Property: get an empty _string.
	static mstring empty() // getter
	{
		mstring result;
		result._cptr = Allocator.reference("").ptr; // Empty, but NOT null!
		assert(result._cptr);
		return result;
	}
	
	
	/// For each character. Modifying the inout dchar in the foreach body may be done in-place or may cause reallocation.
	int opApply(int delegate(inout size_t, inout dchar) dg)
	{
		return _Tapply!(true).each(*this, dg);
	}
	
	/// ditto
	int opApply(int delegate(inout dchar) dg)
	{
		return _Tapply!(false).each(*this, dg);
	}
	
	
	/// Index to a single character.
	dchar opIndex(size_t i)
	{
		switch(_len & TYPE_MASK)
		{
			case TYPE_UTF8:
				return _cptr[0 .. this.length][i];
			case TYPE_UTF16:
				return _wptr[0 .. this.length][i];
			case TYPE_UTF32:
				return _dptr[0 .. this.length][i];
			default: // No type..
				return _cptr[0 .. this.length][i]; // ?
		}
	}
	
	
	/// Assign to a single character. This may be done in-place or may cause reallocation.
	void opIndexAssign(dchar ch, size_t i)
	{
		static if(!(Allocator.FLAGS & MStringAllocatorFlags.MUTABLE))
		{
			assert(0);
		}
		else
		{
			size_t sct;
			switch(_len & TYPE_MASK)
			{
				case TYPE_UTF8:
					sct = smallestCharType(ch);
					if(TYPE_UTF8 == sct)
					{
						_cptr[0 .. this.length][i] = ch;
					}
					else
					{
						switch(sct)
						{
							case TYPE_UTF16:
								{
									wchar[] ws;
									//ws = _mtoUTF16(_cptr[0 .. this.length]);
									ws = _mtoUTFN!(char, wchar)(_cptr[0 .. this.length]);
									Allocator.free(_cptr);
									_wptr = ws.ptr;
									_len = ws.length;
									_len |= TYPE_UTF16;
									_wptr[0 .. this.length][i] = cast(wchar)ch;
								}
								break;
							case TYPE_UTF32:
								{
									dchar[] ds;
									//ds = _mtoUTF32(_cptr[0 .. this.length]);
									ds = _mtoUTFN!(char, dchar)(_cptr[0 .. this.length]);
									Allocator.free(_cptr);
									_dptr = ds.ptr;
									_len = ds.length;
									_len |= TYPE_UTF32;
									_dptr[0 .. this.length][i] = ch;
								}
								break;
							default:
								assert(0);
						}
					}
					break;
				case TYPE_UTF16:
					sct = smallestCharType(ch);
					if(sct <= TYPE_UTF16)
					{
						_wptr[0 .. this.length][i] = cast(wchar)ch;
					}
					else
					{
						dchar[] ds;
						//ds = _mtoUTF32(_wptr[0 .. this.length]);
						ds = _mtoUTFN!(wchar, dchar)(_wptr[0 .. this.length]);
						Allocator.free(_wptr);
						_dptr = ds.ptr;
						_len = ds.length;
						_len |= TYPE_UTF32;
						_dptr[0 .. this.length][i] = ch;
					}
					break;
				case TYPE_UTF32:
					_dptr[0 .. this.length][i] = ch;
					break;
				default: // No type..
					_cptr[0 .. this.length][i] = ch; // ?
			}
			//return ch;
		}
	}
	
	
	/// Get an mstring slice. The returned slice may or may not refer to the same memory, depending on the allocator.
	mstring opSlice(size_t i1, size_t i2)
	in
	{
		assert(i2 >= i1);
	}
	body
	{
		mstring result;
		switch(_len & TYPE_MASK)
		{
			case TYPE_UTF8:
				{
					char[] cs;
					cs = Allocator.reference(_cptr[0 .. this.length][i1 .. i2]);
					result._cptr = cs.ptr;
					result._len = cs.length;
					result._len |= TYPE_UTF8;
				}
				break;
			case TYPE_UTF16:
				{
					wchar[] ws;
					ws = Allocator.reference(_wptr[0 .. this.length][i1 .. i2]);
					result._wptr = ws.ptr;
					result._len = ws.length;
					result._len |= TYPE_UTF16;
				}
				break;
			case TYPE_UTF32:
				{
					dchar[] ds;
					ds = Allocator.reference(_dptr[0 .. this.length][i1 .. i2]);
					result._dptr = ds.ptr;
					result._len = ds.length;
					result._len |= TYPE_UTF32;
				}
				break;
			default: ;
		}
		return result;
	}
	
	/+
	/// ditto
	void[] opSlice()
	{
		switch(_len & TYPE_MASK)
		{
			case TYPE_UTF8:
				return _cptr[0 .. this.length];
			case TYPE_UTF16:
				return _wptr[0 .. this.length];
			case TYPE_UTF32:
				return _dptr[0 .. this.length];
			default: // No type..
				return null;
		}
	}
	+/
	
	
	/// Assign an mstring to a slice. This may be done in-place or may cause reallocation.
	void opSliceAssign(mstring s2, size_t i1, size_t i2)
	{
		static if(!(Allocator.FLAGS & MStringAllocatorFlags.MUTABLE))
		{
			assert(0);
		}
		else
		{
			if(i1 > i2 || i2 > length)
				throw new MStringException("Slice out of bounds");
			if(i2 - i1 != s2.length)
				throw new MStringException("Lengths don't match for string copy");
			
			if((_len & TYPE_MASK) == (s2._len & TYPE_MASK))
			{
				_cptr[i1 .. i2] = s2._cptr[0 .. s2.length];
			}
			else if((_len & TYPE_MASK) > (s2._len & TYPE_MASK))
			{
				switch(_len & TYPE_MASK)
				{
					case TYPE_UTF16:
						switch(s2._len & TYPE_MASK)
						{
							case TYPE_UTF8:
								foreach(wchar wch; s2._cptr[0 .. s2.length])
								{
									_wptr[i1] = wch;
								}
								assert(i1 == i2);
								break;
							default: ;
						}
						break;
					case TYPE_UTF32:
						switch(s2._len & TYPE_MASK)
						{
							case TYPE_UTF8:
								foreach(wchar dch; s2._cptr[0 .. s2.length])
								{
									_dptr[i1] = dch;
								}
								assert(i1 == i2);
								break;
							case TYPE_UTF16:
								foreach(wchar dch; s2._wptr[0 .. s2.length])
								{
									_dptr[i1] = dch;
								}
								assert(i1 == i2);
								break;
							default: ;
						}
						break;
					default: ;
				}
			}
			else //if((_len & TYPE_MASK) < (s2._len & TYPE_MASK))
			{
				switch(s2._len & TYPE_MASK)
				{
					case TYPE_UTF16:
						{
							wchar[] ws;
							//ws = new wchar[this.length];
							ws = Allocator.alloc!(wchar)(this.length);
							size_t iw;
							switch(_len & TYPE_MASK)
							{
								case TYPE_UTF8:
									for(iw = 0; iw != i1; iw++)
									{
										ws[iw] = cast(wchar)_cptr[iw];
									}
									ws[i1 .. i2] = s2._wptr[0 .. s2.length];
									for(iw = i2; iw != ws.length; iw++)
									{
										ws[iw] = cast(wchar)_cptr[iw];
									}
									break;
								default:
									assert(0);
							}
							Allocator.free(_wptr);
							_wptr = ws.ptr;
							_len = ws.length;
							_len |= TYPE_UTF16;
						}
						break;
					case TYPE_UTF32:
						{
							dchar[] ds;
							//ds = new dchar[this.length];
							ds = Allocator.alloc!(dchar)(this.length);
							size_t iw;
							switch(_len & TYPE_MASK)
							{
								case TYPE_UTF8:
									for(iw = 0; iw != i1; iw++)
									{
										ds[iw] = cast(wchar)_cptr[iw];
									}
									ds[i1 .. i2] = s2._dptr[0 .. s2.length];
									for(iw = i2; iw != ds.length; iw++)
									{
										ds[iw] = cast(wchar)_cptr[iw];
									}
									break;
								case TYPE_UTF16:
									for(iw = 0; iw != i1; iw++)
									{
										ds[iw] = cast(wchar)_wptr[iw];
									}
									ds[i1 .. i2] = s2._dptr[0 .. s2.length];
									for(iw = i2; iw != ds.length; iw++)
									{
										ds[iw] = cast(wchar)_wptr[iw];
									}
									break;
								default:
									assert(0);
							}
							Allocator.free(_dptr);
							_dptr = ds.ptr;
							_len = ds.length;
							_len |= TYPE_UTF32;
						}
						break;
					default:
						assert(0);
				}
			}
		}
	}
	
	
	/// Assign a repeating character to a slice. This may be done in-place or may cause reallocation.
	void opSliceAssign(dchar ch, size_t i1, size_t i2)
	{
		static if(!(Allocator.FLAGS & MStringAllocatorFlags.MUTABLE))
		{
			assert(0);
		}
		else
		{
			size_t sct;
			switch(_len & TYPE_MASK)
			{
				case TYPE_UTF8:
					sct = smallestCharType(ch);
					if(TYPE_UTF8 == sct)
					{
						_cptr[0 .. this.length][i1 .. i2] = ch;
					}
					else
					{
						switch(sct)
						{
							case TYPE_UTF16:
								{
									wchar[] ws;
									//ws = _mtoUTF16(_cptr[0 .. this.length]);
									ws = _mtoUTFN!(char, wchar)(_cptr[0 .. this.length]);
									Allocator.free(_cptr);
									_wptr = ws.ptr;
									_len = ws.length;
									_len |= TYPE_UTF16;
									_wptr[0 .. this.length][i1 .. i2] = cast(wchar)ch;
								}
								break;
							case TYPE_UTF32:
								{
									dchar[] ds;
									//ds = _mtoUTF32(_cptr[0 .. this.length]);
									ds = _mtoUTFN!(char, dchar)(_cptr[0 .. this.length]);
									Allocator.free(_cptr);
									_dptr = ds.ptr;
									_len = ds.length;
									_len |= TYPE_UTF32;
									_dptr[0 .. this.length][i1 .. i2] = ch;
								}
								break;
							default:
								assert(0);
						}
					}
					break;
				case TYPE_UTF16:
					sct = smallestCharType(ch);
					if(sct <= TYPE_UTF16)
					{
						_wptr[0 .. this.length][i1 .. i2] = cast(wchar)ch;
					}
					else
					{
						dchar[] ds;
						//ds = _mtoUTF32(_wptr[0 .. this.length]);
						ds = _mtoUTFN!(wchar, dchar)(_wptr[0 .. this.length]);
						Allocator.free(_wptr);
						_dptr = ds.ptr;
						_len = ds.length;
						_len |= TYPE_UTF32;
						_dptr[0 .. this.length][i1 .. i2] = ch;
					}
					break;
				case TYPE_UTF32:
					_dptr[0 .. this.length][i1 .. i2] = ch;
					break;
				default: // No type..
					_cptr[0 .. this.length][i1 .. i2] = ch; // ?
			}
		}
	}
	
	/// ditto
	void opSliceAssign(dchar ch)
	{
		static if(!(Allocator.FLAGS & MStringAllocatorFlags.MUTABLE))
		{
			assert(0);
		}
		else
		{
			return opSliceAssign(ch, 0, length);
		}
	}
	
	
	/// Convert this string to a UTF-8 string of chars; directly returns toUTF8(). Note that this returns normal D memory, not the _mstring allocator.
	char[] toString()
	{
		return toUTF8();
	}
	
	
	/// Concatenation; always returns a copy.
	mstring opCat(mstring s2)
	{
		return _TCat!(true).concat(*this, s2);
	}
	
	/// ditto
	mstring opCat(char[] s2)
	{
		// To-do: optimize.
		mstring mstr2;
		auto r = mstr2._setr(s2, false);
		auto result = _TCat!(true).concat(*this, mstr2);
		mstr2._unsetr(r);
		return result;
	}
	
	/// ditto
	mstring opCat(wchar[] s2)
	{
		// To-do: optimize.
		mstring mstr2;
		auto r = mstr2._setr(s2, false);
		auto result = _TCat!(true).concat(*this, mstr2);
		mstr2._unsetr(r);
		return result;
	}
	
	/// ditto
	mstring opCat(dchar[] s2)
	{
		// To-do: optimize.
		mstring mstr2;
		auto r = mstr2._setr(s2, false);
		auto result = _TCat!(true).concat(*this, mstr2);
		mstr2._unsetr(r);
		return result;
	}
	
	/+
	/// ditto
	string opCat_r(mstring s1)
	{
		return s1.opCat(*this);
	}
	+/
	
	/// ditto
	mstring opCat_r(char[] s1)
	{
		return mstring(_addTo(_mdup(s1))); // Can't use fromASCII.
	}
	
	/// ditto
	mstring opCat_r(wchar[] s1)
	{
		return mstring(_addTo(_mdup(s1))); // Can't use fromBMP.
	}
	
	/// ditto
	mstring opCat_r(dchar[] s1)
	{
		return mstring(_addTo(_mdup(s1)));
	}
	
	
	/// Concatenate and assign. This may be done in-place using adjacent memory or may cause reallocation.
	mstring opCatAssign(mstring s2)
	{
		static if(!(Allocator.FLAGS & MStringAllocatorFlags.MUTABLE))
		{
			assert(0);
		}
		else
		{
			return *this = _TCat!(false).concat(*this, s2); // ...
		}
	}
	
	/// ditto
	mstring opCatAssign(char[] s2)
	{
		static if(!(Allocator.FLAGS & MStringAllocatorFlags.MUTABLE))
		{
			assert(0);
		}
		else
		{
			// To-do: optimize.
			mstring mstr2;
			auto r = mstr2._setr(s2, false);
			opCatAssign(mstr2);
			mstr2._unsetr(r);
			return *this;
		}
	}
	
	/// ditto
	mstring opCatAssign(wchar[] s2)
	{
		static if(!(Allocator.FLAGS & MStringAllocatorFlags.MUTABLE))
		{
			assert(0);
		}
		else
		{
			// To-do: optimize.
			mstring mstr2;
			auto r = mstr2._setr(s2, false);
			opCatAssign(mstr2);
			mstr2._unsetr(r);
			return *this;
		}
	}
	
	/// ditto
	mstring opCatAssign(dchar[] s2)
	{
		static if(!(Allocator.FLAGS & MStringAllocatorFlags.MUTABLE))
		{
			assert(0);
		}
		else
		{
			// To-do: optimize.
			mstring mstr2;
			auto r = mstr2._setr(s2, false);
			opCatAssign(mstr2);
			mstr2._unsetr(r);
			return *this;
		}
	}
	
	/// ditto
	mstring opCatAssign(dchar ch)
	{
		static if(!(Allocator.FLAGS & MStringAllocatorFlags.MUTABLE))
		{
			assert(0);
		}
		else
		{
			if(length + 1 > MAX_LENGTH)
				errortoolong();
			
			size_t sct;
			sct = smallestCharType(ch);
			if(sct <= (_len & TYPE_MASK))
			{
				switch(_len & TYPE_MASK)
				{
					case TYPE_UTF8:
						{
							char[] cs;
							cs = _cptr[0 .. this.length];
							//cs ~= cast(char)ch;
							char ch8 = cast(char)ch;
							cs = Allocator.concat!(char)(cs, (&ch8)[0 .. 1]);
							_cptr = cs.ptr;
							_len = cs.length | TYPE_UTF8;
						}
						break;
					case TYPE_UTF16:
						{
							wchar[] ws;
							ws = _wptr[0 .. this.length];
							//ws ~= cast(wchar)ch;
							wchar ch16 = cast(wchar)ch;
							ws = Allocator.concat!(wchar)(ws, (&ch16)[0 .. 1]);
							_wptr = ws.ptr;
							_len = ws.length | TYPE_UTF16;
						}
						break;
					case TYPE_UTF32:
						{
							dchar[] ds;
							ds = _dptr[0 .. this.length];
							//ds ~= ch;
							ds = Allocator.concat!(dchar)(ds, (&ch)[0 .. 1]);
							_dptr = ds.ptr;
							_len = ds.length | TYPE_UTF32;
						}
						break;
					default:
						assert(0);
				}
				return *this;
			}
			else
			{
				return opCatAssign((&ch)[0 .. 1]);
			}
		}
	}
	
	
	/**
		Create a new instance of string based on an ASCII-only _string.
		This is faster than string(s) (opCall).
		The _string passed in may be used as-is, and any string modifying functions may modify the original _string memory.
	**/
	static mstring fromASCII(char[] s)
	in
	{
		size_t iw;
		for(iw = 0; iw != s.length; iw++)
		{
			assert(!(s[iw] & 0x80)); // String must be within ASCII.
		}
	}
	body
	{
		if(s.length > MAX_LENGTH)
			errortoolong();
		
		mstring result;
		s = Allocator.reference(s);
		result._cptr = s.ptr;
		result._len = s.length;
		result._len |= TYPE_UTF8;
		return result;
	}
	
	/// ditto
	static mstring fromASCII(ubyte[] s)
	{
		return fromASCII(cast(char[])s);
	}
	
	/// ditto
	static mstring fromASCII(byte[] s)
	{
		return fromASCII(cast(char[])s);
	}
	
	
	/**
		Create a new instance of string based on an BMP-Unicode-only wide-character _string.
		This is faster than string(s) (opCall).
		The _string passed in may be used as-is, and any string modifying functions may modify the original _string memory.
		Information on BMP can be found at $(LINK http://en.wikipedia.org/wiki/Basic_Multilingual_Plane)
	**/
	static mstring fromBMP(wchar[] s)
	in
	{
		version(Tango)
		{
		}
		else
		{
			size_t iw;
			for(iw = 0; iw != s.length; iw++)
			{
				size_t idx = 0;
				std.utf.decode(s[iw .. s.length], idx);
				assert(1 == idx); // String must be within BMP.
			}
		}
	}
	body
	{
		if(s.length > MAX_LENGTH)
			errortoolong();
		
		mstring result;
		s = Allocator.reference(s);
		result._wptr = s.ptr;
		result._len = s.length;
		result._len |= TYPE_UTF16;
		return result;
	}
	
	
	/// Convert the string to the different UTF types. Note that the default behavior returns normal D memory, not the _mstring allocator; set useAllocator to true to use the _mstring allocator.
	char[] toUTF8(bool useAllocator = false)
	/+out(result)
	{
		std.utf.validate(result);
	}
	body+/
	{
		switch(_len & TYPE_MASK)
		{
			case TYPE_UTF8:
				if(useAllocator)
					return Allocator.reference(_cptr[0 .. this.length]);
				return _externref(_cptr[0 .. this.length]);
			case TYPE_UTF16:
				//return _mtoUTF8(_wptr[0 .. this.length]);
				if(useAllocator)
					return _mtoUTFN!(wchar, char)(_wptr[0 .. this.length]);
				return _mtoUTFN!(wchar, char, MStringDGcAllocator)(_wptr[0 .. this.length]);
			case TYPE_UTF32:
				//return _mtoUTF8(_dptr[0 .. this.length]);
				if(useAllocator)
					return _mtoUTFN!(dchar, char)(_dptr[0 .. this.length]);
				return _mtoUTFN!(dchar, char, MStringDGcAllocator)(_dptr[0 .. this.length]);
			default: // No type..
				return null;
		}
	}
	
	/// ditto
	wchar[] toUTF16(bool useAllocator = false)
	/+out(result)
	{
		std.utf.validate(result);
	}
	body+/
	{
		switch(_len & TYPE_MASK)
		{
			case TYPE_UTF8:
				//return _mtoUTF16(_cptr[0 .. this.length]);
				if(useAllocator)
					return _mtoUTFN!(char, wchar)(_cptr[0 .. this.length]);
				return _mtoUTFN!(char, wchar, MStringDGcAllocator)(_cptr[0 .. this.length]);
			case TYPE_UTF16:
				if(useAllocator)
					return Allocator.reference(_wptr[0 .. this.length]);
				return _externref(_wptr[0 .. this.length]);
			case TYPE_UTF32:
				//return _mtoUTF16(_dptr[0 .. this.length]);
				if(useAllocator)
					return _mtoUTFN!(dchar, wchar)(_dptr[0 .. this.length]);
				return _mtoUTFN!(dchar, wchar, MStringDGcAllocator)(_dptr[0 .. this.length]);
			default: // No type..
				return null;
		}
	}
	
	/// ditto
	dchar[] toUTF32(bool useAllocator = false)
	/+out(result)
	{
		std.utf.validate(result);
	}
	body+/
	{
		switch(_len & TYPE_MASK)
		{
			case TYPE_UTF8:
				//return _mtoUTF32(_cptr[0 .. this.length]);
				if(useAllocator)
					return _mtoUTFN!(char, dchar)(_cptr[0 .. this.length]);
				return _mtoUTFN!(char, dchar, MStringDGcAllocator)(_cptr[0 .. this.length]);
			case TYPE_UTF16:
				//return _mtoUTF32(_wptr[0 .. this.length]);
				if(useAllocator)
					return _mtoUTFN!(wchar, dchar)(_wptr[0 .. this.length]);
				return _mtoUTFN!(wchar, dchar, MStringDGcAllocator)(_wptr[0 .. this.length]);
			case TYPE_UTF32:
				if(useAllocator)
					return Allocator.reference(_dptr[0 .. this.length]);
				return _externref(_dptr[0 .. this.length]);
			default: // No type..
				return null;
		}
	}
	
	
	/**
		Compare this mstring with another. If ignoreCase is true, 'a' through 'z' are compared case insensitively.
		It is safe to _compare strings with different underlying character UTF types.
		Returns: $(UL $(LI $(B < 0) if this < s2) $(LI $(B = 0) if match) $(LI $(B > 0) if this > s2))
	**/
	int compare(mstring s2, bool ignoreCase = false)
	{
		switch(s2._len & TYPE_MASK)
		{
			case TYPE_UTF8:
				return _scmp(s2._cptr[0 .. s2.length], ignoreCase, true);
			case TYPE_UTF16:
				return _scmp(s2._wptr[0 .. s2.length], ignoreCase, true);
			case TYPE_UTF32:
				return compare(s2._dptr[0 .. s2.length], ignoreCase);
			default:
				assert(!s2._len);
				if(!length)
					return 0;
				return 1;
		}
	}
	
	/// ditto
	int compare(char[] s2, bool ignoreCase = false)
	{
		return _scmp(s2, ignoreCase, false);
	}
	
	/// ditto
	int compare(wchar[] s2, bool ignoreCase = false)
	{
		return _scmp(s2, ignoreCase, false);
	}
	
	/// ditto
	int compare(dchar[] s2, bool ignoreCase = false)
	{
		switch(_len & TYPE_MASK)
		{
			case TYPE_UTF8:
				return _TCmp!(char, dchar, false).compare(_cptr[0 .. this.length], s2, ignoreCase);
			case TYPE_UTF16:
				return _TCmp!(wchar, dchar, false).compare(_wptr[0 .. this.length], s2, ignoreCase);
			case TYPE_UTF32:
				return _TCmp!(dchar, dchar, false).compare(_dptr[0 .. this.length], s2, ignoreCase);
			default:
				assert(!_len);
				if(!s2.length)
					return 0;
				return -1;
		}
	}
	
	
	/// Find a string within this mstring. If ignoreCase is true, 'a' through 'z' are compared case insensitively.
	/// Returns: the starting index of the found string, or -1 if not found.
	int find(mstring s2, bool ignoreCase = false)
	{
		return _TFind!(typeof(s2)).find(*this, s2, ignoreCase);
	}
	
	/// ditto
	int find(char[] s2, bool ignoreCase = false)
	{
		return _TFind!(typeof(s2)).find(*this, s2, ignoreCase);
	}
	
	/// ditto
	int find(wchar[] s2, bool ignoreCase = false)
	{
		return _TFind!(typeof(s2)).find(*this, s2, ignoreCase);
	}
	
	/// ditto
	int find(dchar[] s2, bool ignoreCase = false)
	{
		return _TFind!(typeof(s2)).find(*this, s2, ignoreCase);
	}
	
	
	/// Returns: true if this mstring starts with, or ends with, the provided string. If ignoreCase is true, 'a' through 'z' are compared case insensitively.
	bool startsWith(mstring s, bool ignoreCase = false)
	{
		if(!s.length)
			return true;
		if(s.length > length)
			return false;
		return !(*this)[0 .. s.length].compare(s, ignoreCase);
	}
	
	/// ditto
	bool startsWith(char[] s, bool ignoreCase = false)
	{
		if(!s.length)
			return true;
		if(s.length > length)
			return false;
		return !(*this)[0 .. s.length].compare(s, ignoreCase);
	}
	
	/// ditto
	bool startsWith(wchar[] s, bool ignoreCase = false)
	{
		if(!s.length)
			return true;
		if(s.length > length)
			return false;
		return !(*this)[0 .. s.length].compare(s, ignoreCase);
	}
	
	/// ditto
	bool startsWith(dchar[] s, bool ignoreCase = false)
	{
		if(!s.length)
			return true;
		if(s.length > length)
			return false;
		return !(*this)[0 .. s.length].compare(s, ignoreCase);
	}
	
	/// ditto
	bool endsWith(mstring s, bool ignoreCase = false)
	{
		if(!s.length)
			return true;
		if(s.length > length)
			return false;
		return !(*this)[this.length - s.length .. this.length].compare(s, ignoreCase);
	}
	
	/// ditto
	bool endsWith(char[] s, bool ignoreCase = false)
	{
		if(!s.length)
			return true;
		if(s.length > length)
			return false;
		return !(*this)[this.length - s.length .. this.length].compare(s, ignoreCase);
	}
	
	/// ditto
	bool endsWith(wchar[] s, bool ignoreCase = false)
	{
		if(!s.length)
			return true;
		if(s.length > length)
			return false;
		return !(*this)[this.length - s.length .. this.length].compare(s, ignoreCase);
	}
	
	/// ditto
	bool endsWith(dchar[] s, bool ignoreCase = false)
	{
		if(!s.length)
			return true;
		if(s.length > length)
			return false;
		return !(*this)[this.length - s.length .. this.length].compare(s, ignoreCase);
	}
	
	
	/// Property: get the pointer to the internal _string memory.
	void* ptr() // getter
	{
		return _ptr;
	}
	
	
	/// Returns: true if the pointer to the internal _string memory is null.
	bool isNull()
	{
		//return !_ptr;
		if(_ptr)
			return false;
		assert(!length);
		return true;
	}
	
	
	/// Returns: true if this string has a length of 0, which includes a null mstring.
	bool isEmpty()
	{
		return !length;
	}
	
	
	/// Comparison operators. It is safe to compare strings with different underlying character UTF types.
	int opEquals(mstring s2)
	{
		return !compare(s2, false);
	}
	
	/// ditto
	int opEquals(char[] s2)
	{
		return !compare(s2, false);
	}
	
	/// ditto
	int opEquals(wchar[] s2)
	{
		return !compare(s2, false);
	}
	
	/// ditto
	int opEquals(dchar[] s2)
	{
		return !compare(s2, false);
	}
	
	/// ditto
	int opCmp(mstring s2)
	{
		return compare(s2, false);
	}
	
	/// ditto
	int opCmp(char[] s2)
	{
		return compare(s2, false);
	}
	
	/// ditto
	int opCmp(wchar[] s2)
	{
		return compare(s2, false);
	}
	
	/// ditto
	int opCmp(dchar[] s2)
	{
		return compare(s2, false);
	}
	
	
	/// Reallocates this string using a smaller char type if possible. Generally only needed when slicing.
	mstring compact()
	{
		return *this = _comp(*this, false);
	}
	
	
	/// Returns: duplicate string using new memory; compacts the returned string as necessary.
	mstring dup()
	{
		return _comp(*this, true);
	}
	
	
	/// Release the string memory. Other strings created from this string are still valid, but the underlying memory (ptr) may become invalid.
	void release()
	{
		Allocator.free(_ptr);
		
		_len = 0;
		_ptr = null;
	}
	
	deprecated alias release cleanup;
	
	
	/// The maximum amount of characters allowed in a string.
	const size_t MAX_LENGTH = LENGTH_MASK;
	
	
	private:
	
	size_t _len;
	union
	{
		char* _cptr;
		wchar* _wptr;
		dchar* _dptr;
		void* _ptr;
	}
	
	
	const size_t SIZE_T_BITS = size_t.sizeof * 8;
	
	
	const size_t LENGTH_MASK = (1 << (SIZE_T_BITS - 2)) - 1;
	const size_t TYPE_MASK = ~LENGTH_MASK;
	
	static assert(size_t.sizeof != 4 || mstring.LENGTH_MASK == 0x3FFFFFFF);
	
	
	//const size_t TYPE_UTF8 = 0b01 << (SIZE_T_BITS - 2);
	const size_t TYPE_UTF8 = 0b00;
	const size_t TYPE_UTF16 = 0b10 << (SIZE_T_BITS - 2);
	const size_t TYPE_UTF32 = 0b11 << (SIZE_T_BITS - 2);
	
	static assert(TYPE_UTF32 > TYPE_UTF16 && TYPE_UTF16 > TYPE_UTF8);
	
	
	static size_t smallestCharType(dchar ch)
	{
		return (ch <= 0x7F) ? TYPE_UTF8 : (ch <= 0xFFFF) ? TYPE_UTF16 : TYPE_UTF32;
	}
	
	
	size_t smallestCharTypeForAll()
	{
		size_t result = 0; //TYPE_UTF8; // ?
		foreach(dchar ch; *this)
		{
			size_t x;
			x = smallestCharType(ch);
			if(x > result)
				result = x;
		}
		return result;
	}
	
	
	static void errortoolong()
	{
		throw new MStringException("String length is too long");
	}
	
	
	static mstring _comp(mstring s, bool alwaysCopy)
	{
		if(!s.length)
			return mstring();
		size_t x;
		x = s.smallestCharTypeForAll();
		if(x == (s._len & TYPE_MASK))
		{
			if(alwaysCopy)
			{
				// Can't call s.dup - string.dup calls this.
				//if(s.length)
				{
					switch(x)
					{
						case TYPE_UTF8:
							s._cptr = _mdupFreeOrig(s._cptr[0 .. s.length]).ptr;
							break;
						case TYPE_UTF16:
							s._wptr = _mdupFreeOrig(s._wptr[0 .. s.length]).ptr;
							break;
						case TYPE_UTF32:
							s._dptr = _mdupFreeOrig(s._dptr[0 .. s.length]).ptr;
							break;
						default:
							assert(0);
					}
				}
				/+else
				{
					s = string();
				}+/
			}
			return s;
		}
		assert(x < (s._len & TYPE_MASK));
		size_t iw;
		switch(x)
		{
			case TYPE_UTF32:
				assert(0); // Can never compact to the biggest size.
				return s;
			case TYPE_UTF16:
				{
					wchar[] ws;
					//ws = new wchar[s.length];
					ws = Allocator.alloc!(wchar)(s.length);
					switch(s._len & TYPE_MASK)
					{
						case TYPE_UTF32:
							iw = 0;
							foreach(dchar ch; s._dptr[0 .. ws.length])
							{
								ws[iw++] = cast(wchar)ch;
							}
							Allocator.free(s._dptr); // Free orig.
							assert(iw == ws.length);
							s._wptr = ws.ptr;
							s._len = ws.length;
							s._len |= TYPE_UTF16;
							break;
						default:
							assert(0);
					}
				}
				break;
			case TYPE_UTF8:
					char[] cs;
					//cs = new char[s.length];
					cs = Allocator.alloc!(char)(s.length);
					switch(s._len & TYPE_MASK)
					{
						case TYPE_UTF32:
							iw = 0;
							foreach(dchar ch; s._dptr[0 .. cs.length])
							{
								cs[iw++] = cast(char)ch;
							}
							Allocator.free(s._dptr); // Free orig.
							assert(iw == cs.length);
							s._cptr = cs.ptr;
							s._len = cs.length;
							s._len |= TYPE_UTF8;
							break;
						case TYPE_UTF16:
							iw = 0;
							foreach(wchar ch; s._wptr[0 .. cs.length])
							{
								cs[iw++] = cast(char)ch;
							}
							Allocator.free(s._wptr); // Free orig.
							assert(iw == cs.length);
							s._cptr = cs.ptr;
							s._len = cs.length;
							s._len |= TYPE_UTF8;
							break;
						default:
							assert(0);
					}
				break;
			default:
				assert(!s.length);
				s._len = 0;
				s._ptr = null;
		}
		return s;
	}
	
	
	// -one- gets appended to, possibly in-place.
	T1[] _mappend(T1, T2)(T1[] one, T2[] two)
	{
		T1[] result;
		static if(T1.sizeof == T2.sizeof)
		{
			/+
			size_t idx = one.length;
			result = Allocator.realloc!(T1)(one, one.length + two.length);
			result[idx .. $] = two;
			return result;
			+/
			return Allocator.concat!(T1)(one, two);
		}
		else
		{
			size_t extralen = 0;
			size_t idx = one.length;
			foreach(T1 ch; two)
			{
				extralen++;
			}
			result = Allocator.realloc!(T1)(one, one.length + extralen);
			foreach(T1 ch; two)
			{
				result[idx++] = ch;
			}
			assert(idx == result.length);
			return result;
		}
	}
	
	
	T[] _mappendnew(T)(T[] one, T[] two)
	{
		T[] result = Allocator.alloc!(T)(one.length + two.length);
		result[0 .. one.length] = one;
		result[one.length .. $] = two;
		return result;
	}
	
	
	// Add this string onto s by appending.
	
	char[] _addTo(char[] s)
	{
		size_t thislength;
		thislength = this.length;
		if(!thislength)
			return s;
		switch(_len & TYPE_MASK)
		{
			case TYPE_UTF8:
				//s ~= _cptr[0 .. thislength];
				s = _mappend(s, _cptr[0 .. thislength]);
				break;
			case TYPE_UTF16:
				/+
				foreach(dchar dch; _wptr[0 .. thislength])
				{
					std.utf.encode(s, dch);
				}
				+/
				s = _mappend(s, _wptr[0 .. thislength]);
				break;
			case TYPE_UTF32:
				/+
				foreach(dchar dch; _dptr[0 .. thislength])
				{
					std.utf.encode(s, dch);
				}
				+/
				s = _mappend(s, _dptr[0 .. thislength]);
				break;
			default: ;
		}
		return s;
	}
	
	
	wchar[] _addTo(wchar[] s)
	{
		size_t thislength;
		thislength = this.length;
		if(!thislength)
			return s;
		switch(_len & TYPE_MASK)
		{
			case TYPE_UTF8:
				/+
				foreach(dchar dch; _cptr[0 .. thislength])
				{
					std.utf.encode(s, dch);
				}
				+/
				s = _mappend(s, _cptr[0 .. thislength]);
				break;
			case TYPE_UTF16:
				//s ~= _wptr[0 .. thislength];
				s = _mappend(s, _wptr[0 .. thislength]);
				break;
			case TYPE_UTF32:
				/+
				foreach(dchar dch; _dptr[0 .. thislength])
				{
					std.utf.encode(s, dch);
				}
				+/
				s = _mappend(s, _dptr[0 .. thislength]);
				break;
			default: ;
		}
		return s;
	}
	
	
	dchar[] _addTo(dchar[] s)
	{
		size_t thislength;
		thislength = this.length;
		if(!thislength)
			return s;
		switch(_len & TYPE_MASK)
		{
			case TYPE_UTF8:
				/+
				foreach(dchar dch; _cptr[0 .. thislength])
				{
					std.utf.encode(s, dch);
				}
				+/
				s = _mappend(s, _cptr[0 .. thislength]);
				break;
			case TYPE_UTF16:
				/+
				foreach(dchar dch; _wptr[0 .. thislength])
				{
					std.utf.encode(s, dch);
				}
				+/
				s = _mappend(s, _wptr[0 .. thislength]);
				break;
			case TYPE_UTF32:
				//s ~= _dptr[0 .. thislength];
				s = _mappend(s, _dptr[0 .. thislength]);
				break;
			default: ;
		}
		return s;
	}
	
	
	int _scmp(char[] s2,  bool ignoreCase, bool isASCII)
	{
		switch(_len & TYPE_MASK)
		{
			case TYPE_UTF8:
				if(isASCII)
					return _TCmp!(char, char, false).compare(_cptr[0 .. this.length], s2, ignoreCase);
				return _TCmp!(char, char, true).compare(_cptr[0 .. this.length], s2, ignoreCase);
			case TYPE_UTF16:
				if(isASCII)
					return _TCmp!(wchar, char, false).compare(_wptr[0 .. this.length], s2, ignoreCase);
				return _TCmp!(wchar, char, true).compare(_wptr[0 .. this.length], s2, ignoreCase);
			case TYPE_UTF32:
				if(isASCII)
					return _TCmp!(dchar, char, false).compare(_dptr[0 .. this.length], s2, ignoreCase);
				return _TCmp!(dchar, char, true).compare(_dptr[0 .. this.length], s2, ignoreCase);
			default:
				assert(!_len);
				if(!s2.length)
					return 0;
				return -1;
		}
	}
	
	int _scmp(wchar[] s2,  bool ignoreCase, bool isBMP)
	{
		switch(_len & TYPE_MASK)
		{
			case TYPE_UTF8:
				if(isBMP)
					return _TCmp!(char, wchar, false).compare(_cptr[0 .. this.length], s2, ignoreCase);
				return _TCmp!(char, wchar, true).compare(_cptr[0 .. this.length], s2, ignoreCase);
			case TYPE_UTF16:
				if(isBMP)
					return _TCmp!(wchar, wchar, false).compare(_wptr[0 .. this.length], s2, ignoreCase);
				return _TCmp!(wchar, wchar, true).compare(_wptr[0 .. this.length], s2, ignoreCase);
			case TYPE_UTF32:
				if(isBMP)
					return _TCmp!(dchar, wchar, false).compare(_dptr[0 .. this.length], s2, ignoreCase);
				return _TCmp!(dchar, wchar, true).compare(_dptr[0 .. this.length], s2, ignoreCase);
			default:
				assert(!_len);
				if(!s2.length)
					return 0;
				return -1;
		}
	}
	
	
	template _TFind(TStr)
	{
		int find(mstring s1, TStr s2, bool ignoreCase)
		{
			if(!s2.length)
				return 0;
			if(!s1.length)
				return -1;
			
			size_t psz;
			mstring ts;
			int i, max;
			
			max = s1.length - s2.length + 1;
			ts._ptr = s1._ptr;
			ts._len = s2.length | (s1._len & mstring.TYPE_MASK);
			switch(s1._len & mstring.TYPE_MASK)
			{
				case mstring.TYPE_UTF32: psz = dchar.sizeof; break;
				case mstring.TYPE_UTF16: psz = wchar.sizeof; break;
				case mstring.TYPE_UTF8: psz = char.sizeof; break;
				default: assert(0);
			}
			for(i = 0; i < max; i++)
			{
				if(!ts.compare(s2, ignoreCase))
					return i;
				ts._ptr += psz;
			}
			
			return -1;
		}
	}


	// T1 never UTF converts.
	template _TCmp(T1, T2, bool UTF_CONVERT_T2)
	{
		/+
		dchar nextc(T, bool UTF_CONVERT)(T[] s, inout size_t idx)
		{
			static if(is(T == char) || is(T == wchar) || is(T == dchar))
			{
				static if(UTF_CONVERT)
				{
					version(Tango)
					{
						static assert(0);
					}
					else
					{
						return std.utf.decode(s, idx);
					}
				}
				else
				{
					return cast(dchar)s[idx++];
				}
			}
			else
			{
				static assert(0);
			}
		}
		+/
		dchar nextc(T, bool UTF_CONVERT)(inout T[] s)
		{
			static if(is(T == char) || is(T == wchar) || is(T == dchar))
			{
				static if(UTF_CONVERT)
				{
					version(Tango)
					{
						dchar[1] ds;
						uint ate;
						tango.text.convert.Utf.toString32(s, ds, &ate);
						s = s[ate .. s.length];
						return ds[0];
					}
					else
					{
						size_t idx = 0;
						dchar dch = std.utf.decode(s, idx);
						s = s[idx .. s.length];
						return dch;
					}
				}
				else
				{
					dchar dch = cast(dchar)s[0];
					s = s[1 .. s.length];
					return dch;
				}
			}
			else
			{
				static assert(0);
			}
		}
		
		
		int compare(T1[] s1, T2[] s2, bool ignoreCase)
		{
			size_t max;
			max = s1.length;
			if(s2.length < max)
			{
				if(!s2.length)
					return 1;
				max = s2.length;
			}
			if(!max)
			{
				if(!s2.length)
					return 0;
				return -1;
			}
			size_t iw = 0;
			if(ignoreCase)
			{
				for(;;)
				{
					dchar ch1, ch2, tch1, tch2;
					tch1 = ch1 = nextc!(T1, false)(s1);
					tch2 = ch2 = nextc!(T2, UTF_CONVERT_T2)(s2);
					if(tch1 >= 'A' && tch1 <= 'Z')
						tch1 = cast(dchar)'a' + (tch1 - 'A');
					if(tch2 >= 'A' && tch2 <= 'Z')
						tch2 = cast(dchar)'a' + (tch2 - 'A');
					if(tch1 != tch2)
						return ch1 - ch2;
					if(++iw == max)
						break;
				}
			}
			else
			{
				for(;;)
				{
					dchar ch1, ch2;
					ch1 = nextc!(T1, false)(s1);
					ch2 = nextc!(T2, UTF_CONVERT_T2)(s2);
					if(ch1 != ch2)
						return ch1 - ch2;
					if(++iw == max)
						break;
				}
			}
			//return cast(int)s1.length - cast(int)s2.length;
			return (s1.length == s2.length) ? 0 : (s1.length > s2.length) ? 1 : -1;
		}
	}


	// version MSTRING_ALLOW_FOREACH_MODIFY

	template _Tapply(bool INDEX)
	{
		static if(INDEX)
			alias int delegate(inout size_t, inout dchar) DgType;
		else
			alias int delegate(inout dchar) DgType;
		
		
		int each(inout mstring s, DgType dg)
		{
			size_t iw = 0, len;
			int result = 0;
			len = s.length;
			check_type:
			switch(s._len & mstring.TYPE_MASK)
			{
				case mstring.TYPE_UTF8:
					for(; iw < len; iw++)
					{
						size_t dgi;
						dchar dgch;
						dgi = iw;
						dgch = cast(dchar)s._cptr[iw];
						static if(INDEX)
							result = dg(dgi, dgch);
						else
							result = dg(dgch);
						version(MSTRING_ALLOW_FOREACH_MODIFY)
						{
							if(dgch != cast(dchar)s._cptr[iw])
							{
								s[iw] = dgch; // May cause reallocation.
								if(result)
									break;
								goto check_type;
							}
						}
						else
						{
							assert(dgch == cast(dchar)s._cptr[iw]);
						}
						if(result)
							break;
					}
					break;
				case mstring.TYPE_UTF16:
					for(; iw < len; iw++)
					{
						size_t dgi;
						dchar dgch;
						dgi = iw;
						dgch = cast(dchar)s._wptr[iw];
						static if(INDEX)
							result = dg(dgi, dgch);
						else
							result = dg(dgch);
						version(MSTRING_ALLOW_FOREACH_MODIFY)
						{
							if(dgch != cast(dchar)s._wptr[iw])
							{
								s[iw] = dgch; // May cause reallocation.
								if(result)
									break;
								goto check_type;
							}
						}
						else
						{
							assert(dgch == cast(dchar)s._wptr[iw]);
						}
						if(result)
							break;
					}
					break;
				case mstring.TYPE_UTF32:
					for(; iw < len; iw++)
					{
						size_t dgi;
						dchar dgch;
						dgi = iw;
						dgch = cast(dchar)s._dptr[iw];
						static if(INDEX)
							result = dg(dgi, dgch);
						else
							result = dg(dgch);
						version(MSTRING_ALLOW_FOREACH_MODIFY)
						{
							if(dgch != cast(dchar)s._dptr[iw])
							{
								s[iw] = dgch; // May cause reallocation.
								if(result)
									break;
								goto check_type;
							}
						}
						else
						{
							assert(dgch == cast(dchar)s._dptr[iw]);
						}
						if(result)
							break;
					}
					break;
				default: ;
			}
			return result;
		}
	}
	
	
	static T[] _mdup(T)(T[] arr)
	{
		T[] result;
		result = Allocator.alloc!(T)(arr.length);
		result[] = arr;
		return result;
	}
	
	
	static T[] _mdupFreeOrig(T)(T[] arr)
	{
		T[] result = _mdup(arr);
		Allocator.free(arr.ptr);
		return result;
	}
	
	
	static NT[] _mtoUTFN(T, NT, ToUtfAllocator = Allocator)(T[] s)
	{
		size_t x = 0;
		foreach(NT ch; s)
		{
			x++;
		}
		NT[] result = ToUtfAllocator.alloc!(NT)(x);
		x = 0;
		foreach(NT ch; s)
		{
			result[x++] = ch;
		}
		assert(x == result.length);
		return result;
	}
	
	
	static T[] _externref(T)(T[] a)
	{
		static if(Allocator.FLAGS & MStringAllocatorFlags.REF_EXTERN)
		{
			return a;
		}
		else
		{
			T[] result = MStringDGcAllocator.alloc!(T)(a.length);
			result[] = a;
			return result;
		}
	}


	template _TCat(bool COPY)
	{
		mstring concat(mstring s1, mstring s2)
		{
			size_t newlen;
			newlen = s1.length + s2.length;
			
			if(newlen > mstring.MAX_LENGTH || newlen < s1.length || newlen < s2.length)
				mstring.errortoolong();
			
			mstring result;
			
			if((s1._len & mstring.TYPE_MASK) == (s2._len & mstring.TYPE_MASK))
			{
				switch(s1._len & mstring.TYPE_MASK)
				{
					case mstring.TYPE_UTF8:
						{
							char[] cs;
							cs = array!(char).join(s1._cptr[0 .. s1.length], s2._cptr[0 .. s2.length]);
							result._cptr = cs.ptr;
							result._len = cs.length;
							result._len |= mstring.TYPE_UTF8;
						}
						break;
					case mstring.TYPE_UTF16:
						{
							wchar[] ws;
							ws = array!(wchar).join(s1._wptr[0 .. s1.length], s2._wptr[0 .. s2.length]);
							result._wptr = ws.ptr;
							result._len = ws.length;
							result._len |= mstring.TYPE_UTF16;
						}
						break;
					case mstring.TYPE_UTF32:
						{
							dchar[] ds;
							ds = array!(dchar).join(s1._dptr[0 .. s1.length], s2._dptr[0 .. s2.length]);
							result._dptr = ds.ptr;
							result._len = ds.length;
							result._len |= mstring.TYPE_UTF32;
						}
						break;
					default: ;
				}
			}
			else if((s1._len & mstring.TYPE_MASK) > (s2._len & mstring.TYPE_MASK))
			{
				switch(s1._len & mstring.TYPE_MASK)
				{
					case mstring.TYPE_UTF16:
						{
							wchar[] ws;
							ws = s2._addTo(array!(wchar).get(s1._wptr[0 .. s1.length]));
							result._wptr = ws.ptr;
							result._len = ws.length;
							result._len |= mstring.TYPE_UTF16;
						}
						break;
					case mstring.TYPE_UTF32:
						{
							dchar[] ds;
							ds = s2._addTo(array!(dchar).get(s1._dptr[0 .. s1.length]));
							result._dptr = ds.ptr;
							result._len = ds.length;
							result._len |= mstring.TYPE_UTF32;
						}
						break;
					default:
						assert(0);
				}
			}
			else //if((s1._len & mstring.TYPE_MASK) < (s2._len & mstring.TYPE_MASK))
			{
				switch(s2._len & mstring.TYPE_MASK) // Note: s2's type.
				{
					case mstring.TYPE_UTF16:
						{
							wchar[] ws;
							ws = array!(wchar).join(s1, s2._wptr[0 .. s2.length]);
							result._wptr = ws.ptr;
							result._len = ws.length;
							result._len |= mstring.TYPE_UTF16;
						}
						break;
					case mstring.TYPE_UTF32:
						{
							dchar[] ds;
							ds = array!(dchar).join(s1, s2._dptr[0 .. s2.length]);
							result._dptr = ds.ptr;
							result._len = ds.length;
							result._len |= mstring.TYPE_UTF32;
						}
						break;
					default:
						assert(0);
				}
			}
			
			assert(newlen == result.length);
			return result;
		}
		
		
		// Copies or not based on COPY.
		template array(T)
		{
			T[] join(T[] s1, T[] s2)
			{
				static if(COPY)
					//return s1 ~ s2;
					return _mappendnew(s1, s2);
				else
					//return s1 ~= s2;
					return _mappend(s1, s2);
			}
			
			
			T[] join(mstring s1, T[] s2)
			{
				T[] result;
				result = s1._addTo(cast(T[])null); // Copies.
				//result ~= s2;
				result = _mappend(result, s2);
				return result;
			}
			
			
			T[] get(T[] s)
			{
				static if(COPY)
					return _mdup(s);
				else
					return s;
			}
		}
	}
}


/// 
class MStringException: Exception
{
	///
	this(char[] s)
	{
		super(s);
	}
	
	/+
	/// ditto
	this(mstring s)
	{
		super(s.toString());
	}
	+/
}


/// Flags for an allocator.
enum MStringAllocatorFlags
{
	NONE = 0, ///
	MUTABLE = 0x1, /// The string can be modified.
	REF_EXTERN = 0x2, /// External references (toUTFN(x, false)) don't cause duplication.
}


/** String allocators. An allocator must define all the below static methods and constants.

	Allocators provided:
		MStringDGcAllocator: D's garbage collector allocator (the default);
		MStringCMallocAllocator: C's malloc allocator, need to use release() to prevent leaks;
		MStringMakeAllocatorImmutable: immutable allocator wrapper: MStringMakeAllocatorImmutable(Allocator);
**/
struct MStringAllocator
{
	static:
	
	/// Allocate and return count bytes of type T. Must throw an exception on out-of-memory.
	T[] alloc(T)(size_t count);
	
	/// Reallocate buffer and return count bytes of type T, either in-place or new memory. Must throw an exception on out-of-memory.
	T[] realloc(T)(T[] buffer, size_t count);
	
	/// Concatenate buffer and bufferappend, reallocating (realloc) buffer.
	T[] concat(T)(T[] buffer, T[] bufferappend);
	
	/// Reference the buffer for internal string use. May duplicate or use as-is, depending on the allocator.
	T[] reference(T)(T[] buffer);
	
	/// Release previously allocated or reallocated buffer.
	void free(void* p);
	
	/// Constant specifying this allocator's flags.
	const MStringAllocatorFlags FLAGS;
}


// /// Predefined allocators.
struct MStringDGcAllocator
{
	static:
	
	T[] alloc(T)(size_t count)
	{
		return new T[count];
	}
	
	
	T[] realloc(T)(T[] a, size_t count)
	{
		a.length = count;
		return a;
	}
	
	
	T[] concat(T)(T[] to, T[] arrwith)
	{
		return to ~= arrwith;
	}
	
	
	T[] reference(T)(T[] a)
	{
		return a;
	}
	
	
	void free(void* p)
	{
	}
	
	
	const MStringAllocatorFlags FLAGS = MStringAllocatorFlags.MUTABLE | MStringAllocatorFlags.REF_EXTERN;
}


// /// ditto
struct MStringCMallocAllocator
{
	version(Tango)
	{
		import tango.core.Exception;
		import cstdlib = tango.stdc.stdlib;
		
		class OomException: tango.core.Exception.OutOfMemoryException
		{
			this()
			{
				super(null, 0);
			}
		}
	}
	else
	{
		import std.outofmemory;
		import cstdlib = std.c.stdlib;
		
		alias std.outofmemory.OutOfMemoryException OomException;
	}
	
	static:
	
	T[] alloc(T)(size_t count)
	{
		T* p;
		p = cast(T*)cstdlib.malloc(T.sizeof * count);
		if(!p)
			throw new OomException();
		return p[0 .. count];
	}
	
	
	T[] realloc(T)(T[] a, size_t count)
	{
		T* p;
		p = cast(T*)cstdlib.realloc(a.ptr, T.sizeof * count);
		if(!p)
			throw new OomException();
		return p[0 .. count];
	}
	
	
	T[] concat(T)(T[] to, T[] arrwith)
	{
		size_t oldtolen;
		oldtolen = to.length;
		to = cstdlib.realloc!(T)(to, to.length + arrwith.length);
		to[oldtolen .. to.length] = arrwith;
		return to;
	}
	
	
	T[] reference(T)(T[] a)
	{
		T[] result;
		result = alloc!(T)(a.length);
		result[] = a;
		return result;
	}
	
	
	void free(void* p)
	{
		cstdlib.free(p);
	}
	
	
	const MStringAllocatorFlags FLAGS = MStringAllocatorFlags.MUTABLE;
}


// ///
struct MStringMakeAllocatorImmutable(Allocator)
{
	const MStringAllocatorFlags FLAGS = Allocator.FLAGS & ~MStringAllocatorFlags.MUTABLE;
	
	/+
	alias Allocator.alloc alloc;
	alias Allocator.realloc realloc;
	alias Allocator.concat concat;
	alias Allocator.reference reference;
	alias Allocator.free free;
	+/
	static:
	T[] alloc(T)(size_t count) { return Allocator.alloc!(T)(count); }
	T[] realloc(T)(T[] a, size_t count) { return Allocator.realloc!(T)(a, count); }
	T[] concat(T)(T[] to, T[] arrwith) { return Allocator.concat!(T)(to, arrwith); }
	T[] reference(T)(T[] a) { return Allocator.reference!(T)(a); }
	void free(void* p) { return Allocator.free(p); }
}


/// Default _string type using default D GC allocator.
alias MString!(MStringDGcAllocator) mstring;

version(MSTRING_IMSTRING)
{
	/// An immutable version of mstring.
	alias MString!(MStringMakeAllocatorImmutable!(MStringDGcAllocator)) imstring;
}
else
{
	/// An immutable version of mstring. Must be enabled via version=MSTRING_IMSTRING.
	alias void imstring;
}

version(MSTRING_STRING)
{
	alias mstring string;
}

version(MSTRING_DSTRING)
{
	alias MString DString;
}
else version(MSTRING_NO_DSTRING)
{
}
else
{
	deprecated alias MString DString;
}


//static assert(mstring.sizeof == size_t.sizeof * 2);
static assert(mstring.sizeof == (char[]).sizeof);


version(MSTRING_NO_STRING_EXCEPTION)
{
}
else
{
	deprecated alias MStringException StringException;
}


///
bool isAlphaNumericDiacritic(dchar ch)
{
	return ch >= '\u0300' && ch <= '\u036F';
}


///
bool isSymbolDiacritic(dchar ch)
{
	return ch >= '\u20D0' && ch <= '\u20FF';
}


/// Added to version 4.1 of the Unicode Standard.
bool isSupplimentDiacritic(dchar ch)
{
	return ch >= '\u1DC0' && ch <= '\u1DFF';
}


/// A diacritical mark, or accent mark, combining character.
/// Returns: true if isAlphaNumericDiacritic, isSymbolDiacritic or isSupplimentDiacritic.
bool isDiacritic(dchar ch)
{
	return isAlphaNumericDiacritic(ch) || isSymbolDiacritic(ch) || isSupplimentDiacritic(ch);
}


/// Half mark combining character.
bool isHalfMark(dchar ch)
{
	return ch >= '\uFE20' && ch <= '\uFE2F';
}


/// Includes all of the above diacritics and half marks.
bool isCombiningCharacter(dchar ch)
{
	return isDiacritic(ch) || isHalfMark(ch);
}


///
const REPLACEMENT_CHARACTER = '\uFFFD';


debug(TEST_MAIN)
{
	void main()
	{
		printf("Exiting main.\n");
	}
}


unittest
{
	mstring s = mstring("foo"c);
	assert(s == "foo"c);
	
	assert(s.startsWith("f"c));
	assert(s.startsWith("fo"w));
	assert(s.startsWith("foo"c));
	assert(!s.startsWith("fooo"c));
	assert(!s.startsWith("F"c));
	assert(s.startsWith("F"c, true));
	assert(!s.startsWith("E"c, true));
	assert(s.endsWith("o"c));
	assert(s.endsWith("oo"d));
	assert(s.endsWith("foo"c));
	assert(!s.endsWith("fooo"c));
	assert(!s.endsWith("ooo"c));
	assert(s.endsWith("O"c, true));
	assert(!s.endsWith("P"c, true));
	
	int i;
	s = mstring("The quick brown fox jumped over the lazy dog."w);
	i = s.find("Lazy"c, true);
	assert(-1 != i);
	assert(s[i .. i + 4] == "lazy"c);
	i = s.find("Lazy"c, false);
	assert(-1 == i);
	i = s.find("."c, true);
	assert(-1 != i);
	i = s.find("dog."c, true);
	assert(-1 != i);
	
	wchar[] ws = "Arrrrr!";
	assert(ws == mstring(ws));
	assert(!(ws > mstring(ws)));
	assert(!(ws < mstring(ws)));
	assert(ws == mstring(ws)[0 .. ws.length]);
	assert(!(ws > mstring(ws)[0 .. ws.length]));
	assert(!(ws < mstring(ws)[0 .. ws.length]));
	
	mstring s2;
	s2 = s.dup;
	assert(s.type != s2.type); // Compacted.
	assert(s == s2); // Same value.
	
	/+
	foreach(inout dchar ch; s2)
	{
		//ch = 'z';
		ch = '\u0164';
	}
	+/
	s2[] = '\u0164';
	//printf("`%.*s`\n", s2.toString());
	
	
	// Thanks to Alexander Panek for the following unittests:
	
	  /+string+/ s = mstring("foo"c);
	  mstring t = mstring("Oh noes!"w);
	  mstring u = mstring("% on a plne!"d);
	  mstring v = mstring(mstring(mstring("WAaah! D:"c)[0..2].toUTF16() ~ "arrr! D:"w).toUTF32()); // make some chaos thingie

	  //assert(s.type() == typeid(typeof("foo"c))); // "foo"c is char[3].
	  assert(s.type() != typeid(typeof("Oh noes!"w)));
	  assert(t.type() != typeid(typeof("% on a plne!"d)));
	  assert(v.type() == u.type() );

	  assert(s.startsWith("f"c));
	  assert(s.startsWith("fo"w));
	  assert(s.startsWith("foo"c));
	  assert(!s.startsWith("fooo"c));
	  assert(!s.startsWith("F"c));
	  assert(s.startsWith("F"c, true));
	  assert(!s.startsWith("E"c, true));
	  assert(s.endsWith("o"c));
	  assert(s.endsWith("oo"d));
	  assert(s.endsWith("foo"c));
	  assert(!s.endsWith("fooo"c));
	  assert(!s.endsWith("ooo"c));
	  assert(s.endsWith("O"c, true));
	  assert(!s.endsWith("P"c, true));
	  assert((s.dup ~ t.dup ~ u.dup ~ v.dup).startsWith(s));
	  assert((s.dup ~ t.dup ~ u.dup ~ v.dup).endsWith(v));
	  
	  /+wchar[]+/ ws = "Arrrrr!";
	  
	  assert(!(s == "oof"c));
	  assert(!(t == "oh nes!"w));
	  assert(!(t == u));
	  assert(!(t == s));
	  assert(!(s == u));
	  assert(mstring("foo"c) == s);
	  assert(s == "foo"c);
	  assert(ws == mstring(ws));
	  assert(!(ws > mstring(ws)));
	  assert(!(ws < mstring(ws)));
	  assert(ws == mstring(ws)[0 .. ws.length]);
	  assert(!(ws > mstring(ws)[0 .. ws.length]));
	  assert(!(ws < mstring(ws)[0 .. ws.length]));
	  assert(ws == mstring(ws).toUTF16());
	  //assert(ws == string(mstring(ws).toString()).toUTF16);
	  assert(ws <= (mstring(ws) ~ ws));
	  assert(ws >= mstring(ws)[0..ws.length / 2]);
	  
	  //string s2;
	  s2 = s.dup;
	  //assert(s.type != s2.type); // Compacted.
	  assert(s == s2); // Same value.
	  
	  /+
	  foreach ( uint i, inout dchar ch; u ) {
				 t[i] = ch;
				 //s[s.length - i] = ch;
	  }
	  +/

	  u ~= "\u0164" ~ "\u0020" ~ "\u9876"; // just to be funny and prevent from compacting ;)
	  
	  //assert( t.dup().type( ) == t.type() );
	  u.compact();
	  v.compact();
	  assert( s.type( ) != u.type( ) );
	  
	  assert( s.length < s.MAX_LENGTH );
	  
	  s = mstring();
	  //t.cleanup();
	  t.release();
	  u = mstring();
	  v = mstring();

	  assert(s.length == 0);
	  assert(s.length == t.length);
	  assert(t.length == 0);
	  assert(u.length == 0);
	  assert(v.length == 0);
}

debug(MTEXT_UNITTEST)
{
	import std.stdio;
	void main() { writefln("mtext unittests passed"); }
}
